%option nounput backup nostdinit noyywrap full ecs
%option 8bit backup nodefault perf-report perf-report
%x STRING
%{
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "y.tab.h"

char *text;
static char *text_ptr;
static int text_size, text_asize;

struct buffer {
	struct buffer *parent;
	char *file;
	int lineno;
	YY_BUFFER_STATE state;
};

struct buffer *current_buf;

int lineno(void)
{
	if (current_buf)
		return current_buf->lineno;
	else
		return 0;
}

void scan_init(char *file)
{
	struct buffer *buf = malloc(sizeof(*buf));
	memset(buf, 0, sizeof(*buf));
	buf->file = file;
	buf->lineno = 0;

	yyin = fopen(file, "r");
	if (!yyin) {
		printf("can't find file %s\n", file);
		exit(1);
	}
	current_buf = buf;
}

void scan_nextfile(char *file)
{
	struct buffer *buf = malloc(sizeof(*buf));
	memset(buf, 0, sizeof(*buf));

	buf->parent = current_buf;
	buf->file = file;
	buf->lineno = 0;

	current_buf->state = YY_CURRENT_BUFFER;
	yyin = fopen(file, "r");
	if (!yyin) {
		printf("can't find file %s\n", file);
		exit(1);
	}
	yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
	current_buf = buf;
}

#define START_STRSIZE	16

void new_string(void)
{
	text = malloc(START_STRSIZE);
	text_asize = START_STRSIZE;
	text_ptr = text;
	text_size = 0;
}

void append_string(const char *str, int size)
{
	int new_size = text_size + size + 1;
	if (new_size > text_asize) {
		text = realloc(text, new_size);
		text_asize = new_size;
		text_ptr = text + text_size;
	}
	memcpy(text_ptr, str, size);
	text_ptr += size;
	text_size += size;
	*text_ptr = 0;
}

void alloc_string(const char *str, int size)
{
	text = malloc(size + 1);
	memcpy(text, str, size);
	text[size] = 0;
}

%}
%%
	int str = 0;

\"|\'	{
	str = yytext[0];
	new_string();
	BEGIN(STRING);
}
<STRING>{
	[^'"\n\\]+	{
		append_string(yytext, yyleng);
	}
	\'|\"	{
		if (str == yytext[0]) {
			BEGIN(INITIAL);
			yylval.string = text;
			//printf("s:%s", text);
			return str == '"' ? T_WORD_DQUOTE : T_WORD_QUOTE;
		} else
			append_string(yytext, 1);
	}
	\\[ \t]*\n	append_string(yytext+yyleng-1, 1); current_buf->lineno++;
	\\[ \t]*	append_string(yytext+1, yyleng-1);
	\\.		append_string(yytext+1, 1);
	\n	{
		printf(":%d: open string!\n", current_buf->lineno+1);
		exit(0);
	}
	<<EOF>>	{
		printf(":%d: open string!\n", current_buf->lineno+1);
		exit(0);
	}
}
mainmenu_name[ \t]	return T_MAINMENU_NAME;
mainmenu_option[ \t]	return T_MAINMENU_OPTION;
next_comment/[ \t\n]	return T_NEXT_COMMENT;
endmenu/[ \t\n]		return T_ENDMENU;
comment[ \t]		return T_COMMENT;
text[ \t]		return T_TEXT;
bool[ \t]		return T_BOOL;
hex[ \t]		return T_HEX;
int[ \t]		return T_INT;
string[ \t]		return T_STRING;
tristate[ \t]		return T_TRISTATE;
define_bool[ \t]	return T_DEFINE_BOOL;
define_hex[ \t]		return T_DEFINE_HEX;
define_int[ \t]		return T_DEFINE_INT;
define_string[ \t]	return T_DEFINE_STRING;
define_tristate[ \t]	return T_DEFINE_TRISTATE;
dep_bool[ \t]		return T_DEP_BOOL;
dep_mbool[ \t]		return T_DEP_MBOOL;
dep_hex[ \t]		return T_DEP_HEX;
dep_int[ \t]		return T_DEP_INT;
dep_string[ \t]		return T_DEP_STRING;
dep_tristate[ \t]	return T_DEP_TRISTATE;
unset[ \t]		return T_UNSET;
choice[ \t]		return T_CHOICE;
nchoice[ \t]		return T_NCHOICE;
source[ \t]		return T_SOURCE;
if[ \t]			return T_IF;
else/[ \t\n]		return T_ELSE;
fi/[ \t\n]		return T_FI;
then/[ \t\n]		return T_THEN;
\[			return '[';
\]			return ']';
;			return ';';
!=			return T_UNEQUAL;
=			return '=';
!			return '!';
-a			return T_AND;
-o			return T_OR;
#[^\n]*\n?	{
	/* comment */
	current_buf->lineno++;
	return '\n';
}
[ \t]+			/* space */
\\\n			current_buf->lineno++;	/* ignore */
\n	{
	current_buf->lineno++;
	return '\n';
}
[[:alnum:]$/][[:graph:]]*	{
	alloc_string(yytext, yyleng);
	yylval.string = text;
	//printf("w:%s", text);
	return T_WORD;
}
.	{
		printf(":%d: invalid character 0x%x\n", current_buf->lineno+1, yytext[0]);
		exit(1);
}
<<EOF>>	{
	if (current_buf->parent) {
		current_buf = current_buf->parent;
		yy_delete_buffer(YY_CURRENT_BUFFER);
		yy_switch_to_buffer(current_buf->state);
	} else
		yyterminate();
}
%%
